#!/usr/bin/env python3
#
# Copyright 2020 Alexandre Derumier <aderumier@odiso.com>
# Author: Alexandre Derumier, aderumier@odiso.com
#

try:
    from ifupdown2.lib.addon import Addon

    from ifupdown2.ifupdown.iface import *
    from ifupdown2.ifupdown.utils import utils
    from ifupdown2.ifupdownaddons.modulebase import moduleBase
    from ifupdown2.ifupdown.exceptions import moduleNotSupported
    import ifupdown2.ifupdown.ifupdownflags as ifupdownflags

except Exception:
    from lib.addon import Addon

    from ifupdown.iface import *
    from ifupdown.utils import utils
    from ifupdownaddons.modulebase import moduleBase
    from ifupdown.exceptions import moduleNotSupported
    import ifupdown.ifupdownflags as ifupdownflags

import logging
import re
import subprocess
import os

class openvswitch(Addon, moduleBase):
    """  ifupdown2 addon module to configure Openvswitch bridge """

    _modinfo = {
        'mhelp': 'openvswitch module configure openvswitch bridges',
        'attrs': {
            'ovs-ports': {
                'help': 'Interfaces to be part of this ovs bridge.',
                'validvals': ['<interface-list>'],
                'required': False,
                "multivalue": True,
                "example": [
                    "ovs-ports swp1.100 swp2.100 swp3.100",
                    "ovs-ports glob swp1-3.100",
                    "ovs-ports regex (swp[1|2|3].100)"
                ]
            },
            'ovs-type': {
                'help': 'ovs interface type',
                'validvals': ['OVSBridge'],
                'required': True,
            },
            'ovs-mtu': {
                'help': 'Interface MTU (maximum transmission unit)',
                'validrange': ['552', '9216'],
                'example': ['ovs-mtu 1600'],
                'default': '1500'
            },
            'ovs-options': {
                'help': 'This option lets you add extra arguments to a ovs-vsctl command',
                'required': False,
            },
            'ovs-extra': {
                'help': 'This option lets you run additional ovs-vsctl commands,'  +
                        'separated by "--" (double dash). Variables can be part of the "ovs_extra"' +
                        'option. You can provide all the standard environmental variables' + 
                        'described in the interfaces(5) man page. You can also pass shell' +
                        'commands.extra args',
                'required': False,
                'example': ['ovs_extra set bridge ${IFACE} other-config:hwaddr=00:59:cf:9c:84:3a -- br-set-external-id ${IFACE} bridge-id ${IFACE}']

            },
            'ovs-ports-condone-regex': {
                    "help": "ovs ports to ignore/condone when reloading config / removing interfaces",
                    "required": False,
                    "example": ["ovs-ports-condone-regex ^[a-zA-Z0-9]+_v[0-9]{1,4}$"]
            },
        }
    }

    def __init__ (self, *args, **kargs):
        moduleBase.__init__ (self, *args, **kargs)
        Addon.__init__(self)
        if not os.path.exists('/usr/bin/ovs-vsctl'):
            raise moduleNotSupported('module init failed: no /usr/bin/ovs-vsctl found')

    def _is_ovs_bridge (self, ifaceobj):
        ovstype = ifaceobj.get_attr_value_first('ovs-type')
        if ovstype:
            if ovstype == 'OVSBridge':
                return True
            else:
                return False
        return False

    def _get_ovs_ports (self, ifaceobj):
        ovs_ports = []

        for port in ifaceobj.get_attr_value('ovs-ports') or []:
            ovs_ports.extend(port.split())

        if ovs_ports:
            return self.parse_port_list(ifaceobj.name, ' '.join(ovs_ports))
        else:
            return None

    def _get_running_ovs_ports (self, iface):
        output = utils.exec_command("/usr/bin/ovs-vsctl list-ports %s" %iface)
        if output:
            ovs_ports = sorted(output.splitlines())
            return ovs_ports
        return None

    def _get_ovs_port_condone_regex(self, ifaceobj, get_string = False):
        ovs_port_condone_regex = ifaceobj.get_attr_value_first('ovs-ports-condone-regex')
        if ovs_port_condone_regex:
            if get_string:
                return ovs_port_condone_regex
            return re.compile (r"%s" % ovs_port_condone_regex)
        return None

    def _ovs_vsctl(self, ifaceobj, cmdlist):

        if cmdlist:

            os.environ['IFACE'] = ifaceobj.name if ifaceobj.name else ''
            os.environ['LOGICAL'] = ifaceobj.name if ifaceobj.name else ''
            os.environ['METHOD'] = ifaceobj.addr_method if ifaceobj.addr_method else ''
            os.environ['ADDRFAM'] = ','.join(ifaceobj.addr_family) if ifaceobj.addr_family else ''

            finalcmd = "/usr/bin/ovs-vsctl"

            for cmd in cmdlist:
                finalcmd = finalcmd + " -- " + cmd

            try:
                self.logger.debug ("Running %s" % (finalcmd))
                utils.exec_user_command(finalcmd)
            except subprocess.CalledProcessError as c:
                raise Exception ("Command \"%s failed: %s" % (finalcmd, c.output))
            except Exception as e:
                raise Exception ("%s" % e)

    def _addbridge (self, ifaceobj):

        iface = ifaceobj.name
        ovsoptions = ifaceobj.get_attr_value_first ('ovs-options')
        ovsextra = ifaceobj.get_attr_value('ovs-extra')
        ovsmtu = ifaceobj.get_attr_value_first ('ovs-mtu')
        ovsportscondoneregex = self._get_ovs_port_condone_regex(ifaceobj)

        cmd_list = []

        cmd = "--may-exist add-br %s"%(iface)
        cmd_list.append(cmd)

        if ovsoptions:
            cmd = "set bridge %s %s" %(iface, ovsoptions)
            cmd_list.append(cmd)

        #update
        if self.cache.link_exists (iface):
            # on update, delete active ports not in the new port list
            ovs_ports = self._get_ovs_ports(ifaceobj)
            running_ovs_ports = self._get_running_ovs_ports(iface)

            missingports = []
            if running_ovs_ports is not None and ovs_ports is not None:
                missingports = list(set(running_ovs_ports) - set(ovs_ports))

            if missingports is not None:
                for port in missingports:
                    if ovsportscondoneregex and ovsportscondoneregex.match(port):
                        self.logger.info("%s: port %s will stay enslaved as it matches with ovs-ports-condone-regex" % (ifaceobj.name, port))
                        continue
                    cmd = "--if-exists del-port %s %s"%(iface, port)
                    cmd_list.append(cmd)

            #clear old bridge options
            cmd = "--if-exists clear bridge %s auto_attach controller external-ids fail_mode flood_vlans ipfix mirrors netflow other_config protocols sflow"%(iface)

            cmd_list.append(cmd)

            #clear old interface options
            cmd = "--if-exists clear interface %s mtu_request external-ids other_config options"%(iface)
            cmd_list.append(cmd)

        if ovsextra is not None:
            cmd_list.extend(ovsextra)

        if ovsmtu is not None:
            cmd = "set Interface %s mtu_request=%s"%(iface, ovsmtu)
            cmd_list.append(cmd)

        self._ovs_vsctl(ifaceobj, cmd_list)
        if not self.cache.link_exists(ifaceobj.name):
            self.iproute2.link_add_openvswitch(ifaceobj.name, "openvswitch")

    def _delbridge (self, ifaceobj):

        cmd = "del-br %s"%(ifaceobj.name)
        self._ovs_vsctl(ifaceobj, [cmd])

    def get_dependent_ifacenames (self, ifaceobj, ifaceobjs_all=None):
        if not self._is_ovs_bridge(ifaceobj):
            return None
        ifaceobj.link_privflags |= ifaceLinkPrivFlags.OPENVSWITCH
        return None

    def _up (self, ifaceobj):
        self._addbridge (ifaceobj)

    def _down (self, ifaceobj):
        if not ifupdownflags.flags.PERFMODE and not self.cache.link_exists (ifaceobj.name):
           return

        self._delbridge (ifaceobj)

    def _query_check (self, ifaceobj, ifaceobjcurr):
        if not self.cache.link_exists (ifaceobj.name):
            return
        return

    _run_ops = {
        'pre-up': _up,
        'post-down': _down,
        'query-checkcurr': _query_check
    }

    def get_ops (self):
        """ returns list of ops supported by this module """
        return self._run_ops.keys ()

    def run (self, ifaceobj, operation, query_ifaceobj = None, **extra_args):
        """ run openvswitch configuration on the interface object passed as argument

        Args:
            **ifaceobj** (object): iface object

            **operation** (str): any of 'pre-up', 'post-down', 'query-checkcurr',
                                 'query-running'
        Kwargs:
            **query_ifaceobj** (object): query check ifaceobject. This is only
                valid when op is 'query-checkcurr'. It is an object same as
                ifaceobj, but contains running attribute values and its config
                status. The modules can use it to return queried running state
                of interfaces. status is success if the running state is same
                as user required state in ifaceobj. error otherwise.
        """
        op_handler = self._run_ops.get (operation)
        if not op_handler:
            return

        if (operation != 'query-running' and not self._is_ovs_bridge (ifaceobj)):
            return

        if operation == 'query-checkcurr':
            op_handler (self, ifaceobj, query_ifaceobj)
        else:
            op_handler (self, ifaceobj)



# Copyright (C) 2016, 2017, 2018, 2019 Cumulus Networks, Inc. all rights reserved
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# https://www.gnu.org/licenses/gpl-2.0-standalone.html
#
# Author:
#       Julien Fortin, julien@cumulusnetworks.com
#

import os
import sys
import traceback

import logging
import logging.handlers

root_logger = logging.getLogger()


class LogManager:
    LOGGER_NAME = "ifupdown2"
    LOGGER_NAME_DAEMON = "ifupdown2d"

    DEFAULT_TCP_LOGGING_PORT = 42422
    DEFAULT_LOGGING_LEVEL_DAEMON = logging.INFO
    DEFAULT_LOGGING_LEVEL_NORMAL = logging.WARNING

    __instance = None

    @staticmethod
    def get_instance():
        if not LogManager.__instance:
            try:
                LogManager.__instance = LogManager()
            except Exception as e:
                sys.stderr.write("warning: ifupdown2.Log: %s\n" % str(e))
                traceback.print_exc()
        return LogManager.__instance

    def __init__(self):
        """
        Setup root logger and console handler (stderr). To enable daemon, client
        or standalone logging please call the proper function, see:
            "start_(daemon|client|standlone)_logging"
        """
        if LogManager.__instance:
            raise RuntimeError("Log: invalid access. Please use Log.getInstance()")
        else:
            LogManager.__instance = self

        self.__fmt = "%(levelname)s: %(message)s"

        self.__debug_fmt = "%(asctime)s: %(threadName)s: %(name)s: " \
                           "%(filename)s:%(lineno)d:%(funcName)s(): " \
                           "%(levelname)s: %(message)s"

        self.__root_logger = logging.getLogger()
        self.__root_logger.name = self.LOGGER_NAME

        self.__socket_handler = None
        self.__syslog_handler = None
        self.__console_handler = None

        self.daemon = None

        # by default we attach a console handler that logs on stderr
        # the daemon can manually remove this handler on startup
        self.__console_handler = logging.StreamHandler(sys.stderr)
        self.__console_handler.setFormatter(logging.Formatter(self.__fmt))
        self.__root_logger.addHandler(self.__console_handler)

        if os.path.exists("/dev/log"):
            try:
                self.__syslog_handler = logging.handlers.SysLogHandler(
                    address="/dev/log",
                    facility=logging.handlers.SysLogHandler.LOG_DAEMON
                )
                self.__syslog_handler.setFormatter(logging.Formatter(self.__fmt))
            except Exception as e:
                sys.stderr.write("warning: syslog: %s\n" % str(e))
                self.__syslog_handler = None

        logging.addLevelName(logging.CRITICAL, "critical")
        logging.addLevelName(logging.WARNING, "warning")
        logging.addLevelName(logging.ERROR, "error")
        logging.addLevelName(logging.DEBUG, "debug")
        logging.addLevelName(logging.INFO, "info")

    def set_level(self, default, error=False, warning=False, info=False, debug=False):
        """
        Set root handler logging level
        :param default:
        :param error:
        :param warning:
        :param info:
        :param debug:
        """
        if debug:
            log_level = logging.DEBUG
        elif info:
            log_level = logging.INFO
        elif warning:
            log_level = logging.WARNING
        elif error:
            log_level = logging.ERROR
        else:
            log_level = default

        for handler in self.__root_logger.handlers:
            handler.setLevel(log_level)

        # make sure that the root logger has the lowest logging level possible
        # otherwise some messages might not go through
        if self.__root_logger.level > log_level:
            self.__root_logger.setLevel(log_level)

    def enable_console(self):
        """ Add console handler to root logger """
        self.__root_logger.addHandler(self.__console_handler)

    def disable_console(self):
        """ Remove console handler from root logger """
        self.__root_logger.removeHandler(self.__console_handler)

    def enable_syslog(self):
        """ Add syslog handler to root logger """
        if self.__syslog_handler and self.__syslog_handler not in self.__root_logger.handlers:
            self.__root_logger.addHandler(self.__syslog_handler)

    def disable_syslog(self):
        """ Remove syslog handler from root logger """
        if self.__syslog_handler:
            self.__root_logger.removeHandler(self.__syslog_handler)

    def is_syslog_enabled(self):
        return self.__syslog_handler in self.__root_logger.handlers

    def get_syslog_log_level(self):
        return self.__syslog_handler.level if self.__syslog_handler else None

    def set_level_syslog(self, level):
        if self.__syslog_handler:
            self.__syslog_handler.setLevel(level)

            if self.__root_logger.level > level:
                self.__root_logger.setLevel(level)

    def close_log_stream(self):
        """ Close socket to disconnect client.
        We first have to perform this little hack: it seems like the socket is
        not opened until data (LogRecord) are transmitted. In our most basic use
        case (client sends "ifup -a") the daemon doesn't send back any LogRecord
        but we can't predict that in the client. The client is already in a
        blocking-select waiting for data on it's socket handler
        (StreamRequestHandler). For this special case we need to manually call
        "createSocket" to open the channel to the client so that we can properly
        close it. That way the client can exit cleanly.
        """
        self.__root_logger.removeHandler(self.__socket_handler)
        self.__socket_handler.acquire()
        self.__socket_handler.retryTime = None
        try:
            if not self.__socket_handler.sock:
                self.__socket_handler.createSocket()
        finally:
            self.__socket_handler.close()
            self.__socket_handler.release()

    def start_stream(self):
        self.__root_logger.addHandler(self.__socket_handler)

    def set_daemon_logging_level(self, args):
        self.set_level(self.DEFAULT_LOGGING_LEVEL_DAEMON, info=args.verbose, debug=args.debug)

    def set_request_logging_level(self, args):
        if not hasattr(args, "syslog") or not args.syslog:
            self.disable_syslog()
        else:
            self.__root_logger.removeHandler(self.__socket_handler)
        self.set_level(self.DEFAULT_LOGGING_LEVEL_NORMAL, info=args.verbose, debug=args.debug)

    def start_client_logging(self, args):
        """ Setup root logger name and client log level
        syslog is handled by the daemon directly
        """
        self.__root_logger.name = self.LOGGER_NAME

        if hasattr(args, "syslog") and args.syslog:
            self.enable_syslog()
            self.disable_console()

        self.set_level(self.DEFAULT_LOGGING_LEVEL_NORMAL, info=args.verbose, debug=args.debug)

    def start_standalone_logging(self, args):
        self.__root_logger.name = self.LOGGER_NAME

        if hasattr(args, "syslog") and args.syslog:
            self.enable_syslog()
            self.disable_console()

            self.__root_logger.removeHandler(self.__console_handler)

        self.set_level(self.DEFAULT_LOGGING_LEVEL_NORMAL, info=args.verbose, debug=args.debug)

    def start_daemon_logging(self, args):
        """
        Daemon mode initialize a socket handler to transmit logging to the
        client, we can also do syslog logging and/or console logging (probably
        just for debugging purpose)
        :param args:
        :return:
        """
        self.__root_logger.name = self.LOGGER_NAME_DAEMON
        self.daemon = True

        self.enable_syslog()

        # Create SocketHandler for daemon-client communication
        self.__socket_handler = logging.handlers.SocketHandler(
            "localhost",
            port=self.DEFAULT_TCP_LOGGING_PORT
        )
        self.__root_logger.addHandler(self.__socket_handler)

        if not args.console:
            self.disable_console()

        self.set_daemon_logging_level(args)

    def write(self, msg):
        root_logger.info(msg)

    def root_logger(self):
        return self.__root_logger



# Copyright (C) 2017, 2018 Cumulus Networks, Inc. all rights reserved
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# https://www.gnu.org/licenses/gpl-2.0-standalone.html
#
# Author:
#       Julien Fortin, julien@cumulusnetworks.com
#
# addon -- Addon base class
#

import logging

from collections import OrderedDict

try:
    from ifupdown2.lib.io import IO
    from ifupdown2.lib.sysfs import Sysfs
    from ifupdown2.lib.iproute2 import IPRoute2
    from ifupdown2.lib.base_objects import Netlink, Cache, Requirements
except (ImportError, ModuleNotFoundError):
    from lib.io import IO
    from lib.sysfs import Sysfs
    from lib.iproute2 import IPRoute2
    from lib.base_objects import Netlink, Cache, Requirements


class Addon(Netlink, Cache):
    """
    Base class for ifupdown2 addon modules
    Provides common infrastructure methods for all addon modules
    """

    def __init__(self):
        Netlink.__init__(self)
        Cache.__init__(self)

        self.logger = logging.getLogger("ifupdown2.addons.%s" % self.__class__.__name__)

        self.io = IO()
        self.sysfs = Sysfs
        self.iproute2 = IPRoute2()
        self.requirements = Requirements()

        self.__alias_to_attribute = {}

        for attribute_name, attribute_object in self.__get_modinfo().get("attrs", {}).items():
            for alias in attribute_object.get("aliases", []):
                self.__alias_to_attribute[alias] = attribute_name

    def __get_modinfo(self) -> dict:
        try:
            return self._modinfo
        except AttributeError:
            return {}

    def translate(self, ifaceobjs):
        """
        Replace attribute aliases from user configuration with real attribute name
        """
        for ifaceobj in ifaceobjs:
            ifaceobj.config = OrderedDict(
                [
                    (self.__alias_to_attribute[user_attr], user_value)
                    if user_attr in self.__alias_to_attribute
                    else (user_attr, user_value)
                    for user_attr, user_value in ifaceobj.config.items()
                ]
            )



#!/usr/bin/python3
#
# Copyright 2019 Voleatech GmbH. All rights reserved.
# Author: Sven Auhagen, sven.auhagen@voleatech.de
#

import os
import glob
import socket

try:
    from ifupdown2.lib.addon import Addon

    from ifupdown2.ifupdown.iface import *
    from ifupdown2.ifupdown.utils import utils

    from ifupdown2.nlmanager.nlpacket import Link

    from ifupdown2.ifupdownaddons.modulebase import moduleBase

    import ifupdown2.ifupdown.statemanager as statemanager
    import ifupdown2.ifupdown.policymanager as policymanager
    import ifupdown2.ifupdown.ifupdownflags as ifupdownflags
    import ifupdown2.ifupdown.ifupdownconfig as ifupdownconfig
except (ImportError, ModuleNotFoundError):
    from lib.addon import Addon

    from ifupdown.iface import *
    from ifupdown.utils import utils

    from nlmanager.nlpacket import Link

    from ifupdownaddons.modulebase import moduleBase

    import ifupdown.statemanager as statemanager
    import ifupdown.policymanager as policymanager
    import ifupdown.ifupdownflags as ifupdownflags
    import ifupdown.ifupdownconfig as ifupdownconfig


class xfrm(Addon, moduleBase):
    """
    ifupdown2 addon module to create a xfrm interface
    """
    _modinfo = {
        'mhelp': 'xfrm module creates a xfrm interface for',
        'attrs': {
            'xfrm-id': {
                'help': 'xfrm id',
                'validrange': ['1', '65535'],
                'example': ['xfrm-id 1']
            },
            'xfrm-physdev': {
                'help': 'xfrm physical device',
                'example': ['xfrm-physdev lo']
            },
        },
    }

    def __init__(self, *args, **kargs):
        Addon.__init__(self)
        moduleBase.__init__(self, *args, **kargs)

    def get_dependent_ifacenames(self, ifaceobj, ifacenames_all=None):

        parent_int = self._get_parent_ifacename(ifaceobj)
        if parent_int:
            return [parent_int]

        return None

    def _get_parent_ifacename(self, ifaceobj):
        if ifaceobj.get_attr_value('xfrm-physdev'):
            av_attr = ifaceobj.get_attr_value_first('xfrm-physdev')
            return av_attr

        return None

    def _get_xfrmid(self, ifaceobj):
        if ifaceobj.get_attr_value('xfrm-id'):
            av_attr = ifaceobj.get_attr_value_first('xfrm-id')
            return av_attr

        return None

    def _get_xfrm_name(self, ifaceobj):
        return ifaceobj.name

    @staticmethod
    def _is_my_interface(ifaceobj):
        return ifaceobj.get_attr_value_first('xfrm-id')

    def _up(self, ifaceobj):
        """
        Up the XFRM Interface
        """
        # Create a xfrm device on this device and set the virtual
        # router mac and ip on it
        xfrm_ifacename = self._get_xfrm_name(ifaceobj)
        physdev = self._get_parent_ifacename(ifaceobj)
        xfrmid = self._get_xfrmid(ifaceobj)
        if not self.cache.link_exists(xfrm_ifacename):
            self.iproute2.link_add_xfrm(physdev, xfrm_ifacename, xfrmid)
        else:
            xfrmid_cur = str(
                self.cache.get_link_info_data_attribute(
                    xfrm_ifacename,
                    Link.IFLA_XFRM_IF_ID,
                    0
                )
            )
            physdev_cur = self.cache.get_ifname(
                self.cache.get_link_info_data_attribute(
                    xfrm_ifacename,
                    Link.IFLA_XFRM_LINK,
                    0
                )
            )

            # Check XFRM Values
            if xfrmid != xfrmid_cur or physdev != physdev_cur:
                # Delete and recreate
                self.netlink.link_del(xfrm_ifacename)
                self.iproute2.link_add_xfrm(physdev, xfrm_ifacename, xfrmid)

    def _down(self, ifaceobj, ifaceobj_getfunc=None):
        """
        Down the XFRM Interface
        """
        try:
            xfrm_ifacename = self._get_xfrm_name(ifaceobj)
            self.netlink.link_del(xfrm_ifacename)
        except Exception as e:
            self.log_warn(str(e))

    def _query_check(self, ifaceobj, ifaceobjcurr):
        if not self.cache.link_exists(ifaceobj.name):
            return

        ifaceobjcurr.status = ifaceStatus.SUCCESS

    def _query_running(self, ifaceobjrunning):
        if not self.cache.link_exists(ifaceobjrunning.name):
            return

    # Operations supported by this addon (yet).
    _run_ops = {
        'pre-up': _up,
        'post-down': _down,
        'query-checkcurr': _query_check,
        'query-running': _query_running,
    }

    def get_ops(self):
        return self._run_ops.keys()

    def run(self, ifaceobj, operation, query_ifaceobj=None, **extra_args):
        op_handler = self._run_ops.get(operation)

        if not op_handler:
            return

        if operation != 'query-running' and not self._is_my_interface(ifaceobj):
            return

        if operation == 'query-checkcurr':
            op_handler(self, ifaceobj, query_ifaceobj)
        else:
            op_handler(self, ifaceobj)



# Copyright (C) 2017, 2018 Cumulus Networks, Inc. all rights reserved
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# https://www.gnu.org/licenses/gpl-2.0-standalone.html
#
# Author:
#       Julien Fortin, julien@cumulusnetworks.com
#
# sysfs -- contains all sysfs related operation
#

import os
import glob

try:
    from ifupdown2.lib.io import IO
    from ifupdown2.lib.base_objects import Requirements

    from ifupdown2.ifupdown.utils import utils

    from ifupdown2.nlmanager.nlpacket import Link
except (ImportError, ModuleNotFoundError):
    from lib.io import IO
    from lib.base_objects import Requirements

    from ifupdown.utils import utils

    from nlmanager.nlpacket import Link


class __Sysfs(IO, Requirements):

    __bond_netlink_to_sysfs_attr_map = {
        Link.IFLA_BOND_MODE: "mode",
        Link.IFLA_BOND_MIIMON: "miimon",
        Link.IFLA_BOND_USE_CARRIER: "use_carrier",
        Link.IFLA_BOND_AD_LACP_RATE: "lacp_rate",
        Link.IFLA_BOND_XMIT_HASH_POLICY: "xmit_hash_policy",
        Link.IFLA_BOND_MIN_LINKS: "min_links",
        Link.IFLA_BOND_NUM_PEER_NOTIF: "num_grat_arp",
        Link.IFLA_BOND_AD_ACTOR_SYSTEM: "ad_actor_system",
        Link.IFLA_BOND_AD_ACTOR_SYS_PRIO: "ad_actor_sys_prio",
        Link.IFLA_BOND_AD_LACP_BYPASS: "lacp_bypass",
        Link.IFLA_BOND_UPDELAY: "updelay",
        Link.IFLA_BOND_DOWNDELAY: "downdelay",
        Link.IFLA_BOND_PRIMARY: "primary",
    }

    def __init__(self):
        IO.__init__(self)
        Requirements.__init__(self)

        # Temporary work around to solve the circular dependency with nlcache.
        # Once nlcache is created it will populate sysfs.cache
        self.cache = None

        # if bridge utils is not installed overrrides specific functions to
        # avoid constantly checking bridge_utils_is_installed
        if not Requirements.bridge_utils_is_installed:
            self.bridge_get_mcqv4src = self.bridge_get_mcqv4src_dry_run

    @staticmethod
    def link_get_uppers(ifname):
        try:
            uppers = glob.glob("/sys/class/net/%s/upper_*" % ifname)
            if not uppers:
                return []
            return [os.path.basename(u)[6:] for u in uppers]
        except Exception:
            return []

    @staticmethod
    def link_get_lowers(ifname):
        try:
            lowers = glob.glob("/sys/class/net/%s/lower_*" % ifname)
            if not lowers:
                return []
            return [os.path.basename(l)[6:] for l in lowers]
        except Exception:
            return []

    def link_is_up(self, ifname):
        """
        Read sysfs operstate file
        """
        return "up" == self.read_file_oneline("/sys/class/net/%s/operstate" % ifname)

    def get_link_address(self, ifname):
        """
        Read MAC hardware address from sysfs
        """
        return self.read_file_oneline("/sys/class/net/%s/address" % ifname)

    #
    # MTU
    #

    def link_get_mtu(self, ifname):
        return int(self.read_file_oneline("/sys/class/net/%s/mtu" % ifname) or 0)

    def link_set_mtu(self, ifname, mtu_str, mtu_int):
        if self.cache.get_link_mtu(ifname) != mtu_int:
            if self.write_to_file('/sys/class/net/%s/mtu' % ifname, mtu_str):
                self.cache.override_link_mtu(ifname, mtu_int)

    def link_set_mtu_dry_run(self, ifname, mtu_str, mtu_int):
        # we can remove the cache check in DRYRUN mode
        self.write_to_file('/sys/class/net/%s/mtu' % ifname, mtu_str)

    #
    # ALIAS
    #

    def link_set_alias(self, ifname, alias):
        cached_alias = self.cache.get_link_alias(ifname)

        if cached_alias == alias:
            return

        if not alias:
            alias = "\n"

        if self.write_to_file("/sys/class/net/%s/ifalias" % ifname, alias):
            pass # self.cache.override_link_mtu(ifname, mtu_int)

    def link_set_alias_dry_run(self, ifname, alias):
        # we can remove the cache check in DRYRUN mode
        if not alias:
            alias = ""
        self.write_to_file("/sys/class/net/%s/ifalias" % ifname, alias)

    ############################################################################
    # BRIDGE
    ############################################################################

    def bridge_port_pvids_get(self, bridge_port_name):
        return self.read_file_oneline("/sys/class/net/%s/brport/pvid" % bridge_port_name)

    def bridge_get_stp(self, bridge):
        stp_state_path = "/sys/class/net/%s/bridge/stp_state" % bridge

        if not os.path.exists(stp_state_path):
            return "error"

        stp_state = self.read_file_oneline(stp_state_path)

        if not stp_state:
            return "error"

        try:
            stp_state_int = int(stp_state)
            return "yes" if stp_state_int > 0 else "no"
        except Exception:
            return "unknown"

    def bridge_get_mcqv4src(self, bridge):
        mcqv4src = {}
        try:
            filename = "/sys/class/net/%s/bridge/multicast_v4_queriers" % bridge
            if os.path.exists(filename):
                for line in self.read_file(filename) or []:
                    vlan_id, ip = line.split('=')
                    mcqv4src[vlan_id] = ip.strip()
            return mcqv4src
        except Exception as e:
            self.logger.info("%s showmcqv4src: skipping unsupported command" % utils.brctl_cmd)
            self.bridge_get_mcqv4src = self.bridge_get_mcqv4src_dry_run
            return {}

    @staticmethod
    def bridge_get_mcqv4src_dry_run(bridge):
        return {}

    ############################################################################
    # BOND
    ############################################################################

    def bond_remove_slave(self, bond_name, slave_name):
        if self.cache.is_link_enslaved_to(slave_name, bond_name):
            if self.write_to_file("/sys/class/net/%s/bonding/slaves" % bond_name, "-%s" % slave_name):
                # success we can manually update our cache to make sure we stay up-to-date
                self.cache.override_cache_unslave_link(slave=slave_name, master=bond_name)

    def bond_remove_slave_dry_run(self, bond_name, slave_name):
        self.write_to_file("/sys/class/net/%s/bonding/slaves" % bond_name, "-%s" % slave_name)

    ###

    def bond_create(self, bond_name):
        if self.cache.bond_exists(bond_name):
            return
        self.write_to_file("/sys/class/net/bonding_masters", "+%s" % bond_name)

    def bond_create_dry_run(self, bond_name):
        self.write_to_file("/sys/class/net/bonding_masters", "+%s" % bond_name)

    ###

    def bond_set_attrs_nl(self, bond_name, ifla_info_data):
        """
        bond_set_attrs_nl doesn't need a _dry_run handler because each
        entry in ifla_info_data was checked against the cache already.
        Here write_to_file already has a dry_run handler.
        :param bond_name:
        :param ifla_info_data:
        :return:
        """
        bond_attr_name = 'None'  # for log purpose (in case an exception raised)

        for nl_attr, value in list(ifla_info_data.items()):
            try:
                bond_attr_name = self.__bond_netlink_to_sysfs_attr_map.get(nl_attr)

                if bond_attr_name is None:
                    self.logger.warning(
                        "%s: sysfs configuration: unknown bond attribute %s (value %s)"
                        % (bond_name, nl_attr, value)
                    )
                    continue

                file_path = "/sys/class/net/%s/bonding/%s" % (bond_name, bond_attr_name)
                if os.path.exists(file_path):
                    self.write_to_file(file_path, str(value))
            except Exception as e:
                self.logger.warning("%s: %s %s: %s" % (bond_name, bond_attr_name, value, str(e)))

    ############################################################################
    # /proc/sys/ipv6/conf
    ############################################################################

    def get_ipv6_conf_disable_ipv6(self, ifname):
        return int(self.read_file_oneline("/proc/sys/net/ipv6/conf/%s/disable_ipv6" % ifname) or 0)


Sysfs = __Sysfs()
